/***
*		ornament.c -- ornamentation tool for b&p
*		 1997 Michael Rees (aka Another Option)   All rights reserved.
***/

/*
=============================================================================
|   DATE   | WHAT
=============================================================================
 13 Jan 97 | created
 20 Jan 97 | commented out reading/writing from track clip stuff; added file
           | interface, branching ornament, paste clip
 21 Jan 97 | sysex & snapshot; branching ornaments
 22 Jan 97 | realtime operation now works (i.e. for events with no duration)
 24 Jan 97 | scaling was fucked, but now it's all better; fixed a STUPID 
           | mistake in file writing (delete debug statements, not real stuff);
			  | added event headers and count sizes to savesize()
 17 Oct 97 | started adding real-time controllers
 19 Oct 97 | fleshed real-time controllers ui
 20 Oct 97 | real controller stuff
 21 Oct 97 | updated sysex and load/save routines; snagged potential bug in load()
           | that would be an infinite loop if the size of the saved data was
			  | wrong
*/

#define DBUG 0

//#define FILEDUMP 666

#include <exec/types.h>
#include <exec/memory.h>
#include <dos/dos.h>
#include <intuition/intuition.h>
#include <string.h>

#include <proto/exec.h>
#include <proto/intuition.h>
#include <proto/graphics.h>
#include <clib/alib_protos.h>
#include <clib/alib_stdio_protos.h>
//#include <stdio.h>
#include <stdarg.h>

#include "bars.h"
#include "BarsMacros.h"
#include "BPLib.h"
#include "configlib.h"
#include "ornamentw.h"
#include "proto/ornament_protos.h"
#include "ToolID.h"
#include "/icons/ornamentdown.h"
#include "/icons/ornamentup.h"
#include "snapshotgadget.h"
#include "configgadget2.h"
#include "debug.h"

static UBYTE const versiontag[] = "$VER: Ornament 40.2 (21.10.97)";

#define CTRL_COUNT 7
/* range for ignore & delay gadgets */
#define MAX_IGNORE 768
#define MAX_DELAY  768
/* +/- range for transposition gadget */
#define MAX_TRANSPOSITION 24


struct OrnaTool {
	struct Tool tool;
	struct Event *notes; // lists of stuff to ornament with
	struct Event *cchg;
	struct Event *mtouch;
	struct Event *ptouch;
	struct Event *pbend;
	UWORD flags;
	UWORD count;         // number of events in the ornament
	ULONG duration;      // duration of the ornament
	UWORD ignore;        // don't ornament notes with durations less than this
	UWORD delay;         // delay the start of the ornament this much
	UBYTE gate;          // percentage of input note to keep
	BYTE transposition;  // ornament starts this far from note
	UBYTE percent;       // percentage of notes to ornament
	UBYTE moflags;
	long on[128];        // noteon times
	struct ControllerInfo cinfo[CTRL_COUNT];
};

/*** flags ***/
#define TOOL_OFF             1
#define SCALE_ORNAMENT       2 // scrunch ornament into end of original
#define USE_NOTES            4 // what stuff from the clip to stick in
#define USE_CCHG             8
#define USE_MTOUCH        0x10
#define USE_PTOUCH        0x20
#define USE_PBEND         0x40
#define USE_CONTROLS      0x80 // use real-time controllers
#define SAVED           0x0100 // ornament has been copied to file or clip
#define BRANCH_ORNAMENT 0x0200
#define USE_SYSEX       0x0400
#define FREE_SYSEX      0x0800

/*** moflags ***/
#define SHOW_CCNUMS          1

#define TOOLID Orna_TOOLID

//just the flags & parameters
#define SX_LENGTH 16 
//this gets stuffed into a sysex too
#define CINFO_LENGTH (sizeof(struct ControllerInfo) * CTRL_COUNT)

#define TOOL_SIZE (sizeof(struct OrnaTool))
//to (maybe) support loading an old version of the tool
#define OLD_TOOL_SIZE (TOOL_SIZE - CINFO_LENGTH)

//reset the noteon times array
#define ClearOnTable(t) memset((char *)&t->on,-1,sizeof(long)*128);

#define tw tool->tool.window

extern struct IntuitionBase *IntuitionBase;
extern struct GfxBase *GfxBase;
extern struct LayersBase *LayersBase;
extern struct DOSBase *DOSBase;
extern struct ExecBase *SysBase;
extern struct Functions *functions;

/*==========================================================================
* inittool
*=========================================================================*/

static void __regargs inittool(struct OrnaTool *tool)

{
//	UBYTE i = 0;
	tool->flags = USE_NOTES | FREE_SYSEX | USE_CONTROLS;
	tool->moflags = SHOW_CCNUMS;
	tool->percent = 25;
	tool->ignore = 192;
	tool->transposition = 0;
	tool->gate = 80;
	tool->delay = 0;
	tool->duration = 0;
	tool->count = 0;
	ClearOnTable(tool);
	InitControlInfoArray(&tool->cinfo[0],CTRL_COUNT,CITYPE_NODIFF|CITYPE_OFF,
	                     CIFLAG_DISCARD);
/*
	for ( ; i < CTRL_COUNT ; i++ ) {
		tool->cinfo[i].controller = -1; // turn them off initially
	}
*/
	tool->tool.touched |= TOUCH_INIT;
}

/**********************************************/
/* twiddle and redraw config discard gadgets  */

static void __regargs refreshConfigGadgets(struct OrnaTool *tool)
{
	long i;
	struct Gadget *gadget;

	for (i=C_ORNAMENT ; i<=C_TRANSPOSE ; i++) {
		gadget = functions->GetGadget(tw,i+CTRL_COUNT); //respective discard gadget
		if (tool->cinfo[i-C_ORNAMENT].flags & CIFLAG_DISCARD) {
			gadget->Flags |= SELECTED;
		}
		else {
			gadget->Flags &= ~SELECTED;
		}
		functions->RefreshGadget(tool->tool.window,i+CTRL_COUNT);
	}
}

/*************************************/
/* refresh real-time controller pots */

static void __regargs updateConfigPots(struct OrnaTool *tool)
{
	UBYTE i;
	
	for (i=C_ORNAMENT ; i<=C_TRANSPOSE ; i++) {
		functions->ModifyEmbossedProp(tw,i,tool->cinfo[i-C_ORNAMENT].controller+1,
			                           0,0,0,0,0);
		functions->DrawEmbossedProp(tw,i);
		functions->DrawEmbossed(tw,i+GUP);
		functions->DrawEmbossed(tw,i+GDOWN);
	}
}

/*** ids for lists in a file ***/
#define Note_ID 0x4E6F7465
#define Bend_ID 0x42656E64
#define CChg_ID 0x43436867
#define MTch_ID 0x4D546368
#define PTch_ID 0x50546368
// a sort of a kludge id to make it easy to read the controller stuff without
// rewriting the whole load function and scheme
#define CInf_ID 0x43496E66

/*==========================================================================
* copylist
*
*  INPUT: 
* RETURN: 
*
* EFFECT: copies a list and adjusts time, transposition, next tool
*=========================================================================*/

static struct NoteEvent * __regargs copylist(struct NoteEvent *list,
                                             struct OrnaTool *tool,
                                             long timeoffset, BYTE t)
{
	register struct NoteEvent *copy;
	struct NoteEvent *copylist = NULL;
	struct Tool *nexttool;

	if (tool) {
		if (tool->flags & BRANCH_ORNAMENT) {
			if (tool->tool.branch) nexttool = tool->tool.branch;
		}
		else nexttool = tool->tool.next;
	}
	else nexttool = NULL;

	for ( ; list ; list = list->next ) {
		copy = (struct NoteEvent *)functions->dupeevent((struct Event *)list);
		if (copy) {
			copy->tool = nexttool;
			if (tool) {
				if (tool->flags & BRANCH_ORNAMENT) copy->type |= EVENT_BRANCH;
			}
			copy->time += timeoffset;
			copy->value += t;
			while (copy->value > 127) copy->value -=12;
			while (copy->value < 0) copy->value +=12;
			copy->next = copylist;
			copylist = copy;
		}
		else { // not much use in trying again, eh?
			break;
		}
	}
	return(copylist);
}

#define SCALE_MULT 10000

// don't let durations of scaled events be less than this
#define MINIMUM_DURATION 1

/*================ lifted from stretch.c ===================================
* scalelist
*
*  INPUT: pointer to list of Events, scalar
* RETURN: none
*
* EFFECT: multiplies times and durations by scalar
*=========================================================================*/

static void __inline scalelist(struct NoteEvent *list, ULONG scalar, UWORD delay)
{
	register struct NoteEvent *event = list->next; //leave first event time alone
	long x;

#pragma msg 100 ignore push
	if (IsNoteOn(list)) {
		x = IDivU(IMulU(list->duration,scalar),SCALE_MULT);
		list->duration = (x > MINIMUM_DURATION) ? (UWORD)x : MINIMUM_DURATION;
	}
	for ( ; event ; event=event->next ) {
		x = IDivU(IMulU((event->time - list->time),scalar),SCALE_MULT);
		event->time = x + list->time + delay;
		if (IsNoteOn(event)) {
			x = IDivU(IMulU(event->duration,scalar),SCALE_MULT);
			event->duration = (x > MINIMUM_DURATION) ? (UWORD)x : MINIMUM_DURATION;
		}
	}
}
#pragma msg 100 ignore pop
		
/*********************************************/
static struct NoteEvent * __regargs cloneoffs(struct NoteEvent *list)
{
	register struct NoteEvent *off;
	struct NoteEvent *offlist = NULL;

	for ( ; list ; list = list->next ) {
		if (IsNoteOn(list)) {
			if (off = CloneNoteOff(list)) {
				off->next = offlist;
				offlist = off;
			}
		}
	}
	return(offlist);
}

/*==========================================================================
* ornament
*
*  INPUT: pointers to tool and event
* RETURN: list
*
* EFFECT: adds the ornamentation, and adjusts the input event duration
*=========================================================================*/

static struct NoteEvent * __regargs ornament(struct OrnaTool *tool,
                                             struct NoteEvent *event)
{
	struct NoteEvent *copy;
	struct NoteEvent *ornlist = NULL;
	long time;
	long scale;
	BYTE transposition;
	UBYTE i;
	UWORD flags;

	scale = event->duration; // remember the duration to calculate scaling
#pragma msg 100 ignore push
	time = IDivU(IMulU(event->duration,tool->gate),100);
	event->duration = (time > MINIMUM_DURATION) ? (UWORD)time : MINIMUM_DURATION;
	if (tool->flags & SCALE_ORNAMENT) {
		scale = scale - event->duration;
		scale = (scale > 0) ? IDivU(IMulU(scale,SCALE_MULT),tool->duration) : SCALE_MULT;
	}
#pragma msg 100 ignore pop
	time = event->time + event->duration + tool->delay;
	if (tool->transposition) {
		transposition = event->value + tool->transposition;
		if ((tool->flags & USE_NOTES) && tool->notes) {
			if ((tool->flags & USE_PTOUCH) && tool->ptouch) {
				if (tool->notes->time <= tool->ptouch->time) {
					transposition -= tool->notes->byte1;
				}
				else {
					transposition -= tool->ptouch->byte1;
				}
			}
			else {
				transposition -= tool->notes->byte1;
			}
		}
		else if ((tool->flags & USE_PTOUCH) && tool->ptouch) {
			transposition -= tool->ptouch->byte1;
		}
	}
	else transposition = 0;
	for ( i = 0 ; i < 5 ; i++ ) {
		switch (i) {
			case 0 :
				flags = tool->flags & USE_NOTES;
				copy = (struct NoteEvent *)tool->notes;
				break;
			case 1 :
				flags = tool->flags & USE_PTOUCH;
				copy = (struct NoteEvent *)tool->ptouch;
				break;
			case 2 :
				flags = tool->flags & USE_CCHG;
				copy = (struct NoteEvent *)tool->cchg;
				transposition = 0;
				break;
			case 3 :
				flags = tool->flags & USE_MTOUCH;
				copy = (struct NoteEvent *)tool->mtouch;
				break;
			case 4 :
				flags = tool->flags & USE_PBEND;
				copy = (struct NoteEvent *)tool->pbend;
				break;
		}
		if (flags && copy) {
			copy = copylist(copy,tool,time,transposition);
			if (copy) {
				ornlist = (struct NoteEvent *)functions->List_Cat(
				                  (struct ListItem *)ornlist,(struct ListItem *)copy);
			}
		}
	}
	if (ornlist && (tool->flags & USE_NOTES) && tool->notes) {
		copy = cloneoffs(ornlist);
		if (copy) {
			ornlist = (struct NoteEvent *)functions->List_Cat(
			                     (struct ListItem *)ornlist,(struct ListItem *)copy);
		}
	}
	if (ornlist && (tool->flags & SCALE_ORNAMENT)) {
		ornlist = (struct NoteEvent *)
		                          functions->sorteventlist((struct Event *)ornlist);
		scalelist(ornlist,scale,tool->delay);
	}
	return(ornlist);
}

#define SECRET_MESSAGE 9999

/****************************************************************************/
/* note that MEMF_CLEAR in the myalloc() call screws up the Message somehow */
/****************************************************************************/
static void __regargs SendSecretMsg(struct Tool *tool,UWORD secret_message)
{
	struct Message *msg;

	if (msg = (struct Message *)functions->myalloc(sizeof(struct Message),0)) {
		msg->mn_Node.ln_Type = NT_MESSAGE;
		msg->mn_ReplyPort = NULL;
		msg->mn_Length = secret_message;
		PutMsg(tool->window->UserPort,msg);
	}
}

/*==========================================================================
* dosysex
*
*  INPUT: pointer to StringEvent
* RETURN: pointer to Event
*
* EFFECT: gets new tool state from sysex message
*=========================================================================*/

static struct Event * __inline dosysex(struct OrnaTool *tool,
                                       struct StringEvent *event)
{
	register unsigned char *sxdata;
	UWORD sxflag;

	if (tool->flags & USE_SYSEX) {
		if (sxdata = checksysex(event,TOOL_SYSEX_ID,tool->tool.toolid)) {
			sxflag = tool->flags & (USE_SYSEX | FREE_SYSEX);
			memcpy((void *)&tool->flags,(void *)sxdata,SX_LENGTH);
			memcpy((void *)&tool->cinfo,(void *)&sxdata[SX_LENGTH],CINFO_LENGTH);
			if (sxflag & USE_SYSEX) tool->flags |= USE_SYSEX;
			if (sxflag & FREE_SYSEX) tool->flags |= FREE_SYSEX;
			if (tool->tool.window) SendSecretMsg(&tool->tool,SECRET_MESSAGE);
			if (tool->flags & FREE_SYSEX) {
				functions->freestring(event->string);
				functions->freeevent((struct Event *)event);
				event=NULL;
			}
		}
	}
	return((struct Event *)event);
}

/**************************************/

static void __regargs playornament(struct OrnaTool *tool)
{
	struct NoteEvent *outlist = NULL;
	struct NoteEvent *next;
	struct NoteEvent *list;
	struct Tool *outtool;
	UBYTE i;
	UWORD flags;

	if (InToolPad(&tool->tool)) return; //temporarily out of service
	outtool = tool->tool.next;
	while (!(outtool->tooltype & TOOL_OUTPUT)) outtool = outtool->next;
	if (outtool) {
		for ( i = 0 ; i < 5 ; i++ ) {
			switch (i) {
				case 0 :
					flags = tool->flags & USE_NOTES;
					list = (struct NoteEvent *)tool->notes;
					break;
				case 1 :
					flags = tool->flags & USE_CCHG;
					list = (struct NoteEvent *)tool->cchg;
					break;
				case 2 :
					flags = tool->flags & USE_MTOUCH;
					list = (struct NoteEvent *)tool->mtouch;
					break;
				case 3 :
					flags = tool->flags & USE_PTOUCH;
					list = (struct NoteEvent *)tool->ptouch;
					break;
				case 4 :
					flags = tool->flags & USE_PBEND;
					list = (struct NoteEvent *)tool->pbend;
					break;
			}
			if (flags && list) {
				next = copylist(list,NULL,functions->timenow,0);
				if (next) {
					outlist = (struct NoteEvent *)
					                    functions->List_Cat((struct ListItem *)outlist,
				                                           (struct ListItem *)next);
				}
			}
		}
		if ((tool->flags & USE_NOTES) && tool->notes) {
			next = cloneoffs(outlist);
			if (next) {
				functions->List_Cat((struct ListItem *)outlist,
				                    (struct ListItem *)next);
			}
		}
		for ( ; outlist ; outlist = next ) {
			next = outlist->next;
			outlist->tool = outtool;
			functions->qevent((struct Event *)outlist);
		}
	}
}

/************************
 * scale cc value to range defined by min..max
 */

static UWORD __regargs scalecc(UBYTE val, UWORD  min, UWORD max)
{
#pragma msg 100 ignore push
	return((UWORD)(IDivU(IMulU(IDivU(IMulU(max-min+1,100),127),val),100) + min));
#pragma msg 100 pop
}

/************************************************************************/
#define BITEME 0xFADE
#define SECRET_CC (SECRET_MESSAGE - 1)

static struct Event * __regargs docontrolchange(struct OrnaTool *tool,
                                                struct Event *event)
{
	BYTE i = -1;      /* cinfo[] index */
	UBYTE gotone = 0; /* found entry for this cc# */

	if (event->byte1 > 120) return(event); /* none of the midi resets need apply */
//dprintf("docc() cc=%ld b2=%ld\n",event->byte1,event->byte2);
	/* else process controls */
	while ( (i = FindController(&tool->cinfo[0],event->byte1,++i)) > -1 ) {
		gotone=1;
		if (tool->cinfo[i].flags & CIFLAG_DISCARD) event->data = BITEME;
		switch (i) {
			case 0 :  // ornament %
				tool->percent = (UBYTE)scalecc(event->byte2,0,100);
//dprintf("found percent cc\n");
				break;
			case 1 :  // ornament route
				if (event->byte2 < 43) {
					tool->flags = (tool->flags & ~BRANCH_ORNAMENT) | TOOL_OFF;
				}
				else if (event->byte2 < 86) {
					tool->flags &= ~(BRANCH_ORNAMENT | TOOL_OFF);
				}
				else {
					tool->flags = (tool->flags & ~TOOL_OFF) | BRANCH_ORNAMENT;
				}
				break;
			case 2 :  // gate %
				tool->gate = (UBYTE)scalecc(event->byte2,0,100);
				break;
			case 3 :  // scale on/off
				if (SwitchOn(event->byte2)) tool->flags |= SCALE_ORNAMENT;
				else tool->flags &= ~SCALE_ORNAMENT;
				break;
			case 4 :  // delay amount
				tool->delay = scalecc(event->byte2,0,MAX_DELAY);
				break;
			case 5 :  // ignore threshold
				tool->ignore = scalecc(event->byte2,0,MAX_IGNORE);
				break;
			case 6 :  // transposition
				tool->transposition = 
					scalecc(event->byte2,0,MAX_TRANSPOSITION<<1) - MAX_TRANSPOSITION;
				break;
		}
	}
	if (event->data == BITEME) {
		functions->freeevent(event);
		event=NULL;
	}
	if (gotone && tw) SendSecretMsg(&tool->tool,SECRET_CC);
	return(event);
}
		
#define donow(t) ((functions->random(100) <= t->percent) || (t->percent == 100))

/*==========================================================================
* processevent
*=========================================================================*/

static struct Event * __stdargs __saveds processevent(struct NoteEvent *event)

{
   struct OrnaTool *tool = (struct OrnaTool *)event->tool;
	struct NoteEvent *ornamentlist = NULL;
	struct NoteEvent fake;

	if (!tool->tool.touched) inittool(tool);
	event->tool = event->tool->next;
//dprintf("pe() type=%ld b1=%ld b2=%ld\n",event->status,event->value,event->velocity);
	if (IsSysEx(event)) {
		event =(struct NoteEvent *)dosysex(tool,(struct StringEvent *)event);
	}
	else if (IsCtrlChg(event) && (tool->flags & USE_CONTROLS)) {
//dprintf("flags=%ld\n",tool->flags);
		event = (struct NoteEvent *)docontrolchange(tool,(struct Event *)event);
	}
	else if (!(tool->flags & TOOL_OFF) && tool->percent) {
		if (IsNoteOff(event)) {
			if (tool->on[event->value] != -1) {
				if ((event->time - tool->on[event->value]) > tool->ignore) {
					if (donow(tool)) {
						memset((char *)&fake,0,sizeof(struct NoteEvent));
						fake.time = tool->on[event->value];
						fake.type = EVENT_VOICE;
						fake.value = event->value;
						fake.duration = event->time - fake.time;
						fake.tool = event->tool;
						ornamentlist = ornament(tool,&fake);
						event->next = ornamentlist;
						event->time = tool->on[event->value] + fake.duration;
//						event->duration = fake.duration;
						event = (struct NoteEvent *)
						                functions->sorteventlist((struct Event *)event);
					}
				}
				tool->on[event->value] = -1;
			}
/*
			else {
				functions->freeevent((struct Event *)event);
				event = NULL;
			}
*/
		}
		else if (IsNoteOn(event)) {
			if (event->duration < 1) { // this is a live event; ornament on noteoff 
				tool->on[event->value] = event->time;
				return((struct Event *)event);
			}
			/* else */
			if (event->duration >= tool->ignore) {
				if (donow(tool)) {
					ornamentlist = ornament(tool,event);
					event->next = CloneNoteOff(event);
					if (event->next && ornamentlist) {
						event->next->next = ornamentlist;
					}
					else {
						functions->freeevent((struct Event *)event);
						event = ornamentlist;
					}
					event = (struct NoteEvent *)functions->sorteventlist((struct Event *)event);
				}
			}
		}
	}
	return((struct Event *)event);
}

/*==========================================================================
* processlist
*
*  INPUT: pointers to tool and list to process, start and stop times
* RETURN: pointer to inverted list
*
* EFFECT: processes a list of events
*=========================================================================*/

static struct Event * __stdargs __saveds processlist(struct OrnaTool *tool,
                                                     struct NoteEvent *list,
                                                     long start, long stop)
{
	struct NoteEvent *ornamentlist = NULL;
	struct NoteEvent *olist;
	struct NoteEvent *event;

	if (!tool->tool.touched) inittool(tool);
	if (start < 0) start = 0;
	event = list;
	while (event->time < start) {
		event->tool = tool->tool.next;
		event = event->next;
	}
	for ( ; event ; event = event->next ) {
		if (event->time > stop) break;
		event->tool = tool->tool.next;
		if (IsNoteOn(event) && (event->duration >= tool->ignore)) {
			if (donow(tool)) {
				if (olist = ornament(tool,event)) {
					ornamentlist = (struct NoteEvent *)functions->List_Cat(
                        (struct ListItem *)ornamentlist,(struct ListItem *)olist);
				}
			}
		}
	}
	if (ornamentlist) {
		list = (struct NoteEvent *)functions->List_Cat((struct ListItem *)list,
		                                           (struct ListItem *)ornamentlist);
		list = (struct NoteEvent *)functions->sorteventlist((struct Event *)list);
	}
	return((struct Event *)list);
}

/*==========================================================================
* freelists
*
*  INPUT: pointer to tool
* RETURN: none
*
* EFFECT: frees event lists allocated to the tool
*=========================================================================*/

static void __regargs freelists(struct OrnaTool *tool)
{
	if (tool->notes) {
		functions->freelist(tool->notes);
		tool->notes = NULL;
	}
	if (tool->cchg) {
		functions->freelist(tool->cchg);
		tool->cchg = NULL;
	}
	if (tool->mtouch) {
		functions->freelist(tool->mtouch);
		tool->mtouch = NULL;
	}
	if (tool->ptouch) {
		functions->freelist(tool->ptouch);
		tool->ptouch = NULL;
	}
	if (tool->pbend) {
		functions->freelist(tool->pbend);
		tool->pbend = NULL;
	}
	tool->duration = 0;
}

/***************************
 display the duration of and number of events in the current ornament
***************************/

static void __regargs showstats(struct OrnaTool *tool)
{
	struct Gadget *g;
	UBYTE beats;

	if (tw) {
		g = functions->GetGadget(tw,LENGTH);
		if (g) {
#pragma msg 104 ignore push
			sprintf(g->GadgetText->IText,"%4ld",tool->count);
			g->GadgetText->IText[4] = ' ';
			beats = IDivU((UWORD)tool->duration,192);
			sprintf(&g->GadgetText->IText[13],"%02ld.%03ld)",
			                               beats, tool->duration - IMulU(beats,192));
#pragma msg 104 pop
			RefreshGList(g,tw,NULL,1);
		}
	}
}

/*==========================================================================
* copycliplists
*=========================================================================*/

static long __regargs copycliplists(struct OrnaTool *tool,struct Clip *clip,
                                    long begin, long end)
{
	struct Event *event;
	struct Event *next;
	struct Event *last = NULL;
	long timeoffset;
	long x = 0;

	Forbid();
	if (clip->locked) {
		Permit();
#pragma msg 104 ignore push
		functions->msgbox("Can't copy: the clip is locked!",NULL,NULL," Poo ",1,
	                     BP_LGREY,BP_RED,BP_BLUE);
#pragma msg 104 pop
		return(0);
	}
	clip->locked = 1;
	Permit();
	if (clip->events.first && (begin != end)) {
		event = clip->events.first;
		while (event->time < begin) event = event->next;
		if (event) {
			if (event->time > end) event = NULL;
			else {
				timeoffset = event->time;
				freelists(tool);
			}
		}
		/* copy the events in the clip to the appropriate list */
		for (next = event ; next ; next=next->next) {
			if (next->time > end) break;
			if (event = functions->dupeevent(next)) {
				event->time -= timeoffset; // list times start at 0
				switch (event->status) {
					case MIDI_NOTEON :
						x++;
						event->next = tool->notes;
						tool->notes = event;
						break;
					case MIDI_CCHANGE :
						x++;
						event->next = tool->cchg;
						tool->cchg = event;
						break;
					case MIDI_MTOUCH :
						x++;
						event->next = tool->mtouch;
						tool->mtouch = event;
						break;
					case MIDI_PTOUCH :
						x++;
						event->next = tool->ptouch;
						tool->ptouch = event;
						break;
					case MIDI_PBEND :
						x++;
						event->next = tool->pbend;
						tool->pbend = event;
						break;
					default :
						break;
				}
				last = event;
			}
		}
		if (tool->notes) tool->notes = functions->sorteventlist(tool->notes);
		if (tool->cchg) tool->cchg = functions->sorteventlist(tool->cchg);
		if (tool->mtouch) tool->mtouch = functions->sorteventlist(tool->mtouch);
		if (tool->ptouch) tool->ptouch = functions->sorteventlist(tool->ptouch);
		if (tool->pbend) tool->pbend = functions->sorteventlist(tool->pbend);
		if (last) {
			tool->duration = last->time;
			if (IsNoteOn(last)) tool->duration += ((struct NoteEvent *)last)->duration;
		}
		tool->flags &= ~SAVED;
	}
	clip->locked = 0;
	return(x);
}

/*==========================================================================
* writeclip
*
*  INPUT: pointer to tool
* RETURN: yea or nay
*
* EFFECT: copies lists to a clip and puts it to the clipboard
*=========================================================================*/

static long __regargs writeclip(struct OrnaTool *tool)
{
	struct Clip *clip;
	char name[2] = "x";

	if (tool->count) {
		clip = (struct Clip *)functions->myalloc(sizeof(struct Clip),0);
		memset((void *)clip,0,sizeof(struct Clip));
		if (clip) {
			if (tool->notes) {
				clip->events.first = (struct Event *)
			                    copylist((struct NoteEvent *)tool->notes,NULL,0,0);
			}
			if (tool->cchg) {
				clip->events.first = (struct Event *)functions->List_Cat(
			                                   (struct ListItem *)clip->events.first,
           (struct ListItem *)copylist((struct NoteEvent *)tool->cchg,NULL,0,0));
			}
			if (tool->mtouch) {
				clip->events.first = (struct Event *)functions->List_Cat(
			                                   (struct ListItem *)clip->events.first,
           (struct ListItem *)copylist((struct NoteEvent *)tool->mtouch,NULL,0,0));
			}
			if (tool->ptouch) {
				clip->events.first = (struct Event *)functions->List_Cat(
			                                   (struct ListItem *)clip->events.first,
           (struct ListItem *)copylist((struct NoteEvent *)tool->ptouch,NULL,0,0));
			}
			if (tool->pbend) {
				clip->events.first = (struct Event *)functions->List_Cat(
			                                   (struct ListItem *)clip->events.first,
           (struct ListItem *)copylist((struct NoteEvent *)tool->pbend,NULL,0,0));
			}
			clip->events.first = (struct Event *)
			                            functions->sorteventlist(clip->events.first);
			clip->begin = 0;
			clip->end = tool->duration;
			clip->name = functions->allocstring(name);
			functions->clipboard(CLIP_PUT,clip);
			functions->freestring(clip->name);
			functions->freelist(clip->events.first);
			functions->myfree((char *)clip,sizeof(struct Clip));
			tool->flags |= SAVED;
			return(1);
		}
	}
	else return(0);
}


/*==========================================================================
* readclip
*
*  INPUT: pointer to tool
* RETURN: t/f
*
* EFFECT: reads ornament from clipboard, and installs it
*=========================================================================*/

static long __regargs readclip(struct OrnaTool *tool)
{
//	struct Clip clip;  /*? needs to be __far so b&p gets the right address ?*/
	struct Clip *clip; /* but this works */
	long x = 0;

	clip = (struct Clip *)functions->myalloc(sizeof(struct Clip),0);
	if (clip) {
		memset((void *)clip,0,sizeof(struct Clip));
		x = functions->clipboard(CLIP_GET,clip);
		if (x == 1) {
			x = copycliplists(tool,clip,clip->begin,clip->end);
		}
		else x = 0;
		functions->myfree((char *)clip,sizeof(struct Clip));
	}
	return(x);
}


/************** track reading stuff removed 20 jan 97 **********
****************************************************************

/***********************************/
static struct Clip * masterclip(void)

{
	if (functions->masteredit && functions->masteredit->inedit) {
		return(&functions->masteredit->clip);
	}
	else {
		return(&functions->masterclip);
	}
}

/***********************************/
static struct Clip * __regargs trackclip(struct Tool *tool,long *begin,long *end)

{
	struct Clip *clip = NULL;
	struct Tool *t;
	struct Track *track;

	if (InToolPad(tool)) {
		track = functions->tracklist;
		for ( ; track ; track=track->next) {
			if (track == functions->selectedtrack) {
/*
dprintf("track->selected=%ld clip=%ld ",track->selected,&track->clip);
if (track->edit) dprintf("edit->clip=%ld",&track->edit->clip);
*/
				break;
			}
		}
//dprintf("\n");
		if (track) {
			if (track->edit && track->edit->inedit) {
				clip = &track->edit->clip;
				*begin = track->edit->begin; // use edit window markers if editing
				*end = track->edit->end;
			}
			else {
				clip = &track->clip;
				*begin = functions->padcutin; // main markers if not editing
				*end = functions->padcutout;
			}
/*
dprintf("found selected: track=%ld (%s) edit->track=%ld clip=%ld\n",track,
          functions->stringtext(track->clip.name),track->edit->track,clip);
*/
		}
		else clip = NULL;
		if (clip) {
			if (clip == masterclip()) clip = NULL;
		}
	}

	else {
//dprintf("in pipeline:");
		track = functions->tracklist;
		for ( ; track ; track=track->next) {
			for (t=track->toollist ; t ; t=t->next) {
				if (t == tool) break;
			}
			if (t == tool) break;
		}
		if (track) {
			if (track->edit && track->edit->inedit) {
				clip = &track->edit->clip;
				*begin = track->edit->begin; // use edit window markers if editing
				*end = track->edit->end;
			}
			else {
				clip = &track->clip;
				*begin = functions->padcutin; // main markers if not editing
				*end = functions->padcutout;
			}
/*
			if (clip) {
				*begin = track->edit->begin;
				*end = track->edit->end;
			}
*/
//dprintf("  clip=%ld begin=%ld end=%ld\n",clip,*begin,*end);
		}
/*
if (track) {
 dprintf("\nfound me track=%ld name=%s track->bits=%ld\n",
           track,functions->stringtext(track->clip.name),track->bits);
 dprintf("&track->clip=%ld track->clip.begin=%ld .end=%ld .events.first=%ld locked=%ld\n",
   &track->clip,track->clip.begin,track->clip.end,track->clip.events.first,track->clip.locked);
 if (track->edit) {
  dprintf("&track->edit->clip=%ld begin=%ld end=%ld .events=%ld inedit=%ld locked=%ld\n",
            &track->edit->clip,track->edit->begin,track->edit->end,track->edit->clip.events.first,
			   track->edit->inedit,track->edit->clip.locked);
 }
}
if (tool) {
 dprintf("tool->track=%ld &tool->clip=%ld tool->track->edit=%ld\n",
          tool->track,tool->clip,tool->track->edit);
 if (tool->track->edit) {
  dprintf("tool->track->edit->clip.events.first=%ld tool->track->edit->inedit=%ld\n",
           tool->track->edit->clip.events.first,tool->track->edit->inedit);
 }
}
*/
	}

	return(clip);
}

/*==========================================================================
* readtrack
*
*  INPUT: pointer to tool
* RETURN: t/f
*
* EFFECT: gets an ornament from a track, and installs it
*=========================================================================*/

static long __regargs readtrack(struct OrnaTool *tool)
{
	struct Clip *clip;
	long begin;
	long end;
	long x = 0;

	clip = trackclip(&tool->tool,&begin,&end);
//dprintf("readtrack clip=%ld begin=%ld end=%ld\n",clip,begin,end);
	if (clip) {
		x = copycliplists(tool,clip,begin,end);
	}
	return(x);
}

********************************************************
********************************************************/


/*==========================================================================
* dosnapshot
* ?? should this save the event lists too ??
*=========================================================================*/

static void __regargs dosnapshot(struct OrnaTool *tool)

{
	UBYTE sx[SX_LENGTH + CINFO_LENGTH];
	struct StringEvent *event;
	
	memcpy(sx,&tool->flags,SX_LENGTH);
	memcpy(&sx[SX_LENGTH],&tool->cinfo,CINFO_LENGTH);
	if (event = BuildSnapshot(&tool->tool, 
	                          functions->running ? functions->timenow : -1,
	                          sx, SX_LENGTH)) {
		PostSnapshot(&tool->tool,event);
	}
}

/*==========================================================================
* hiloPlusMinus
*
* EFFECT: displays +/-12 for a pot
*=========================================================================*/

static long __stdargs __saveds hiloPlusMinus(struct Window *w,struct Gadget *g,long x)
{
	char text[4];

	Move(w->RPort,g->LeftEdge+g->Width+XOffset+5,g->TopEdge+YOffset);
	SetAPen(w->RPort,1);
	SetBPen(w->RPort,0);
	SetDrMd(w->RPort,JAM2);
	x -= MAX_TRANSPOSITION;
#pragma msg 104 ignore push
	if (x <= 0) sprintf(text,"%3ld",x);
	else sprintf(text,"+%2ld",x);
#pragma msg 104 pop
	Text(w->RPort,text,3);
	return(x);
}

static struct OrnaTool *selectedtool = NULL;
/***
 *** displays time m.bb.ccc, 0=='Nothing', '<' preceding time
 ***/

static long __stdargs __saveds hiloBeatMe(struct Window *w,struct Gadget *g,long x)
{
	char text[13];

	if (x==0) {
		strcpy(text,"   Nothing ");
	}
	else {
		functions->lengthtostring(selectedtool->tool.clip,0,x,text);
		text[2] = '<';
	}
	SetRPort(w->RPort, g->LeftEdge+g->Width+XOffset+5, g->TopEdge+YOffset,
            HILO_APEN,HILO_BPEN, HILO_DRAWMODE);
	Text(w->RPort,&text[2],9);
	return((long)x);
}

/***
 *** displays time m.bb.ccc
 ***/

static long __stdargs __saveds hiloBeat(struct Window *w,struct Gadget *g,long x)
{
	char text[13];

	functions->lengthtostring(selectedtool->tool.clip,0,x,text);
	SetRPort(w->RPort, g->LeftEdge+g->Width+XOffset+13, g->TopEdge+YOffset,
            HILO_APEN,HILO_BPEN, HILO_DRAWMODE);
	Text(w->RPort,&text[3],8);
	return((long)x);
}

static char * const ccnames[] = {
/*0 - 4 */	"MonPres", "ModWhel", "Breath ", "MIDI 03", "Foot   ",
/*5 - 9 */	"PortTim", "DataEnt", "Volume ", "Balance", "MIDI 09",
/*10-14 */	"Pan    ", "Express", "MIDI 12", "MIDI 13", "MIDI 14",
/*15-19 */	"MIDI 15", "Genl 1 ", "Genl 2 ", "Genl 3 ", "Genl 4 ",
/*20-24 */	"MIDI 20", "MIDI 21", "MIDI 22", "MIDI 23", "MIDI 24",
/*25-29 */	"MIDI 25", "MIDI 26", "MIDI 27", "MIDI 28", "MIDI 29",
/*30-34 */	"MIDI 30", "MIDI 31", "MIDI 32", "MIDI 33", "MIDI 34",
/*35-39 */	"MIDI 35", "MIDI 36", "MIDI 37", "MIDI 38", "MIDI 39",
/*40-44 */	"MIDI 40", "MIDI 41", "MIDI 42", "MIDI 43", "MIDI 44",
/*45-49 */	"MIDI 45", "MIDI 46", "MIDI 47", "MIDI 48", "MIDI 49",
/*50-54 */	"MIDI 50", "MIDI 51", "MIDI 52", "MIDI 53", "MIDI 54",
/*55-59 */	"MIDI 55", "MIDI 56", "MIDI 57", "MIDI 58", "MIDI 59",
/*60-64 */	"MIDI 60", "MIDI 61", "MIDI 62", "MIDI 63", "Sustain",
/*65-69 */	"PortSwt", "Sosten.", "SoftPed", "MIDI 68", "FrezPdl",
/*70-74 */	"MIDI 70", "MIDI 71", "MIDI 72", "MIDI 73", "MIDI 74",
/*75-79 */	"LegtoSw", "MIDI 76", "MIDI 77", "MIDI 78", "MIDI 79",
/*80-84 */	"Genl 5 ", "Genl 6 ", "Genl 7 ", "Genl 8 ", "MIDI 84",
/*85-89 */	"MIDI 85", "MIDI 86", "MIDI 87", "MIDI 88", "MIDI 89",
/*90-94 */	"MIDI 90", "FXDepth", "TremDep", "ChorDep", "CelDep ",
/*95-99 */	"PhsrDep", "DataInc", "DataDec", "MIDI 98", "MIDI 99",
/*100-104*/	"MIDI100", "MIDI101", "MIDI102", "MIDI103", "MIDI104",
/*105-109*/	"MIDI105", "MIDI106", "MIDI107", "MIDI108", "MIDI109",
/*110-114*/	"MIDI110", "MIDI111", "MIDI112", "MIDI113", "MIDI114",
/*115-119*/	"MIDI115", "MIDI116", "MIDI117", "MIDI118", "MIDI119",
/*120-124*/	"MIDI120", "RsetCtl" /*, "LoclCtl", "AllOff ", "OmniOff",*/
/*125-127*/	/*"Omni On", "Mono On", "Poly On"*/
};

/* pens for config pot value text */
#define TEXTCOLOR 1
#define BACKCOLOR 0

/* display pot value where 0 is Off */

static long __stdargs __saveds hiloOff(struct Window *w, struct Gadget *g, long h)

{
	char text[8];
	
	Move(w->RPort,g->LeftEdge+g->Width+XOffset,g->TopEdge+YOffset);
	SetAPen(w->RPort,TEXTCOLOR);
	SetBPen(w->RPort,BACKCOLOR);
	SetDrMd(w->RPort,JAM2);
	if (--h > -1) {
		if (selectedtool->moflags & SHOW_CCNUMS) {
#pragma msg 104 ignore push
			sprintf(text,"%3ld",h);
#pragma msg 104 ignore pop
			strcat(&text[3],"    ");
			Text(w->RPort,text,7);
		}
		else {
			Text(w->RPort,ccnames[h],7);
		}
	}
	else {
#pragma msg 104 ignore push
		sprintf(text,"%s","Off");
#pragma msg 104 ignore pop
		strcat(&text[3],"    ");
		Text(w->RPort,text,7);
	}
	return(h);
}


#define REFRESH_STATS         2
#define REFRESH_POPPIPES   0x10
#define REFRESH_CONFIG     0x40
#define REFRESH_DRAWCONFIG 0x80
/*==========================================================================
* edittool
*
*  INPUT: pointer to tool
* RETURN: none
*
* EFFECT: ui handler
*=========================================================================*/

static void __stdargs __saveds edittool(struct OrnaTool *tool)

{
	struct IntuiMessage *message;
	long class; 
	long code;
	struct Gadget *gadget;
	struct NewWindow *nw;
	struct Menu *menu;
	struct MenuItem *menuitem;
	UBYTE refresh;

	if (!(tool->tool.touched & TOUCH_INIT)) inittool(tool);
	if (tool->flags & USE_NOTES) SubItem4.Flags |= CHECKED;
	else SubItem4.Flags &= ~CHECKED;
	if (tool->flags & USE_CCHG) SubItem5.Flags |= CHECKED;
	else SubItem5.Flags &= ~CHECKED;
	if (tool->flags & USE_MTOUCH) SubItem6.Flags |= CHECKED;
	else SubItem6.Flags &= ~CHECKED;
	if (tool->flags & USE_PTOUCH) SubItem7.Flags |= CHECKED;
	else SubItem7.Flags &= ~CHECKED;
	if (tool->flags & USE_PBEND) SubItem8.Flags |= CHECKED;
	else SubItem8.Flags &= ~CHECKED;
	if (tool->flags & USE_CONTROLS) MenuItem10b.Flags |= CHECKED;
	else MenuItem10b.Flags &= ~CHECKED;
	if (tool->moflags & SHOW_CCNUMS) MenuItem10.Flags |= CHECKED;
	else MenuItem10.Flags &= ~CHECKED;
	if (refresh = InToolPad(&tool->tool)) refresh = 0;
	else refresh = 40;
	nw=OpenEmbossedWindowSlop(&NewWindowStructure1,&tool->tool,0,WTITLE,refresh);
	if (!nw) return;
	if (refresh) { // do this if not in Toolpad
		AddConfigGadget(tw);
		AddSnapshotGadget(tw);
		if (tool->flags & USE_SYSEX) SubItem1.Flags |= CHECKED;
		else SubItem1.Flags &= ~CHECKED;
		if (tool->flags & FREE_SYSEX) SubItem2.Flags |= CHECKED;
		else SubItem2.Flags &= ~CHECKED;
		SubItem1.Flags |= ITEMENABLED;
		SubItem2.Flags |= ITEMENABLED;
		SubItem3.Flags |= ITEMENABLED;
	}
	else { // these don't work in the toolpad
		SubItem1.Flags &= ~ITEMENABLED;
		SubItem2.Flags &= ~ITEMENABLED;
		SubItem3.Flags &= ~ITEMENABLED;
		functions->EnableEmbossed(tw,PLAY,0);
	}
	if (menu = functions->DupeMenu(&MenuList1)) SetMenuStrip(tw,menu);
	else SetWindowTitles(tw,(UBYTE *)-1,MenuList1.MenuName);
	PrintIText(tw->RPort,&IntuiTextList1,0,0);
	functions->EmbossOn(tw,SCALE,1);
	functions->EmbossOn(tw,PLAY,1);
	functions->FatEmbossedPropOn(tw,ORNAMENT,ORNAMENT+GDOWN,ORNAMENT+GUP,
	                             hiloPercent,101,1);
	functions->FatEmbossedPropOn(tw,IGNORE,IGNORE+GDOWN,IGNORE+GUP,
	                             hiloBeatMe,MAX_IGNORE+1,1);
	functions->FatEmbossedPropOn(tw,GATE,GATE+GDOWN,GATE+GUP,hiloPercent,101,1);
	functions->FatEmbossedPropOn(tw,TRANSPOSE,TRANSPOSE+GDOWN,TRANSPOSE+GUP,
	                             hiloPlusMinus,(MAX_TRANSPOSITION<<1)+1,1);
	functions->FatEmbossedPropOn(tw,DELAY,DELAY+GDOWN,DELAY+GUP,
	                             hiloBeat,MAX_DELAY+1,1);
	for (code = C_ORNAMENT ; code <= C_TRANSPOSE ; code++) { //emboss config pots
		functions->FatEmbossedPropOn(tw,code,code+GDOWN,code+GUP,hiloOff,122,1);
	}
	if (WindowIsTall(tw)) {
		DrawConfigFrame(tw);
		refresh = REFRESH_CONFIG | REFRESH_DRAWCONFIG;
	}
	else refresh = 0;
	refresh |= (REFRESH_STATS | REFRESH_POPPIPES | 1);

	for ( ; ; ) {
		if (refresh) {
			if (refresh & 1) {
				if (tool->gate == 100) tool->flags &= ~SCALE_ORNAMENT;
				functions->SelectEmbossed(tw,SCALE,tool->flags & SCALE_ORNAMENT);
				if (InToolPad(&tool->tool))
					functions->EnableEmbossed(tw,PLAY,tool->count);
				functions->ModifyEmbossedProp(tw,ORNAMENT,tool->percent,0,0,0,0,0);
				functions->DrawEmbossedProp(tw,ORNAMENT);
				functions->ModifyEmbossedProp(tw,TRANSPOSE,
				                        tool->transposition+MAX_TRANSPOSITION,0,0,0,0,0);
				functions->DrawEmbossedProp(tw,TRANSPOSE);
				functions->ModifyEmbossedProp(tw,GATE,tool->gate,0,0,0,0,0);
				functions->DrawEmbossedProp(tw,GATE);
				selectedtool = tool;
				functions->ModifyEmbossedProp(tw,DELAY,tool->delay,0,0,0,0,0);
				functions->DrawEmbossedProp(tw,DELAY);
				functions->ModifyEmbossedProp(tw,IGNORE,tool->ignore,0,0,0,0,0);
				functions->DrawEmbossedProp(tw,IGNORE);
				RefreshGList(functions->GetGadget(tw,POPUP_IGNORE),tw,NULL,1);
				RefreshGList(functions->GetGadget(tw,POPUP_DELAY),tw,NULL,1);
				if (!InToolPad(&tool->tool)) {
					RefreshGList(functions->GetGadget(tw,SNAPSHOT_GADGET_ID),tw,NULL,1);
				}
// 	      functions->EnableMenuItem(menu,0,SAVEFILE_ITEM,0,tool->count);
				functions->EnableMenuItem(menu,0,WRITECLIP_ITEM,0,tool->count);
			}
			if (refresh & REFRESH_STATS) showstats(tool);
			if (refresh & REFRESH_POPPIPES) {
				gadget = functions->GetGadget(tw,POPPIPE);
				if (tool->flags & TOOL_OFF) {
					gadget->GadgetRender = (APTR)&PlugImage;
				}
				else if (tool->flags & BRANCH_ORNAMENT) {
					gadget->GadgetRender = (APTR)&ElbowImage;
				}
				else {
					gadget->GadgetRender = (APTR)&PipeImage;
				}
				RefreshGList(gadget,tw,NULL,1);
			}
			if (refresh & REFRESH_CONFIG) {
				UpdateControllers(&tool->cinfo[0]);
				updateConfigPots(tool);
			}
			if (refresh & REFRESH_DRAWCONFIG) {
				refreshConfigGadgets(tool);
			}
			refresh = 0;
		}
		message = (struct IntuiMessage *)functions->GetIntuiMessage(tw);
		if (message->ExecMessage.mn_Length == SECRET_MESSAGE)
			refresh = 0xFF;
		else if (message->ExecMessage.mn_Length == SECRET_CC)
			refresh = 0xFF & ~(REFRESH_CONFIG | REFRESH_DRAWCONFIG);
		if (refresh) {
			functions->myfree((char *)message,sizeof(struct Message));
			continue;
		}
		/* else */
		class = message->Class;
		code = message->Code;
		gadget = (struct Gadget *)message->IAddress;
		ReplyMsg((struct Message *)message);
		class = functions->SystemGadgets(tw,class,gadget,code);

		if (class == CLOSEWINDOW) {
			break;
		}

		else if (class == NEWSIZE) {
			if (WindowIsTall(tw)) {
				DrawConfigFrame(tw);
				refresh = REFRESH_CONFIG | REFRESH_DRAWCONFIG;
			}
			SizeTitleBar(tw,-40);
			functions->RefreshGadget(tw,SNAPSHOT_GADGET_ID);
			functions->RefreshGadget(tw,CONFIG_GADGET_ID);
		}

		else if (class == MENUPICK) {
			while (code != MENUNULL) {
				menuitem = ItemAddress(menu,code);
				switch (ITEMNUM(code)) {
					case OPENFILE_ITEM :
						break;
					case SAVEFILE_ITEM:
						break;
					case READCLIP_ITEM :
						if (code = readclip(tool)) {
							tool->count = code;
							refresh = REFRESH_STATS;
						}
#pragma msg 104 ignore push
						else {
							functions->msgbox("Couldn't copy from the clipboard!",
						                  " Drat ",NULL,NULL,1,BP_LGREY,BP_RED,BP_BLUE);
						}
#pragma msg 104 pop
						break;
					case WRITECLIP_ITEM :
						writeclip(tool);
						break;
					case INCLUDE_ITEM :
						switch (SUBNUM(code)) {
							case NOTES_SUBITEM :
								tool->flags ^= USE_NOTES;
								break;
							case CCHG_SUBITEM :
								tool->flags ^= USE_CCHG;
								break;
							case MTOUCH_SUBITEM :
								tool->flags ^= USE_MTOUCH;
								break;
							case PTOUCH_SUBITEM :
								tool->flags ^= USE_PTOUCH;
								break;
							case PBEND_SUBITEM :
								tool->flags ^= USE_PBEND;
								break;
						}
						break;
					case SYSEX_ITEM :
						switch (SUBNUM(code)) {
							case USESYSEX_SUBITEM :
								tool->flags ^= USE_SYSEX;
								break;
							case FREESYSEX_SUBITEM :
								tool->flags ^= FREE_SYSEX;
								break;
							case SNAPSHOT_SUBITEM :
								dosnapshot(tool);
								break;
						}
						break;

					case USECC_ITEM :
						tool->flags ^= USE_CONTROLS;
						break;
					case SHOWCCNUM_ITEM :
						tool->moflags ^= SHOW_CCNUMS;
						updateConfigPots(tool);
						break;

					default: break;
				}
				code = menuitem->NextSelect;
			}
		}

		else if (class == GADGETUP) {
			switch (gadget->GadgetID) {
				case DIS_ORNAMENT :
				case DIS_ROUTE :
				case DIS_GATE :
				case DIS_SCALE :
				case DIS_DELAY :
				case DIS_IGNORE :
				case DIS_TRANSPOSE :
					tool->cinfo[gadget->GadgetID - DIS_ORNAMENT].flags ^= CIFLAG_DISCARD;
					refresh = REFRESH_CONFIG | REFRESH_DRAWCONFIG;
					break;
				case SCALE :
					tool->flags ^= SCALE_ORNAMENT;
					refresh = 1;
					break;
				case PLAY :
					playornament(tool);
					break;
				case SNAPSHOT_GADGET_ID :
					dosnapshot(tool);
					break;
				case CONFIG_GADGET_ID :
					ZoomWindow(tw);
					break;
				default : break;
			}
		}

		else if (class == GADGETDOWN) {
			selectedtool = tool;
			switch (gadget->GadgetID) {
				case C_ORNAMENT :
				case C_ROUTE :
				case C_GATE :
				case C_SCALE :
				case C_DELAY :
				case C_IGNORE :
				case C_TRANSPOSE :
					tool->cinfo[gadget->GadgetID - C_ORNAMENT].controller =
						functions->DragEmbossedProp(tw,gadget->GadgetID);
					refresh = REFRESH_CONFIG;
					break;

				case C_ORNAMENT+GDOWN :
				case C_ROUTE+GDOWN :
				case C_GATE+GDOWN :
				case C_SCALE+GDOWN :
				case C_DELAY+GDOWN :
				case C_IGNORE+GDOWN :
				case C_TRANSPOSE+GDOWN :
					tool->cinfo[gadget->GadgetID - C_ORNAMENT - GDOWN].controller =
						   functions->ShiftEmbossedProp(tw,gadget->GadgetID-GDOWN,-1,0);
					refresh = REFRESH_CONFIG;
					break;

				case C_ORNAMENT+GUP :
				case C_ROUTE+GUP :
				case C_GATE+GUP :
				case C_SCALE+GUP :
				case C_DELAY+GUP :
				case C_IGNORE+GUP :
				case C_TRANSPOSE+GUP :
					tool->cinfo[gadget->GadgetID - C_ORNAMENT - GUP].controller =
						      functions->ShiftEmbossedProp(tw,gadget->GadgetID-GUP,1,0);
					refresh = REFRESH_CONFIG;
					break;

				case POPPIPE :
					code = (short)functions->PopUpMenu(tw,&PopPipeMenu);
					if (ITEMNUM(code) < 3) {
						switch (ITEMNUM(code)) {
							case ELBOW_ITEM :
								tool->flags = (tool->flags & ~TOOL_OFF)|BRANCH_ORNAMENT;
								break;
							case PLUG_ITEM :
								tool->flags = (tool->flags & ~BRANCH_ORNAMENT)|TOOL_OFF;
								break;
							case PIPE_ITEM :
								tool->flags &= ~(TOOL_OFF | BRANCH_ORNAMENT);
								break;
						}
						refresh = REFRESH_POPPIPES;
					}
					break;
				case ORNAMENT :
					tool->percent = functions->DragEmbossedProp(tw,ORNAMENT);
					refresh = 1;
					break;
				case ORNAMENT+GDOWN :
					tool->percent = functions->ShiftEmbossedProp(tw,ORNAMENT,-1,0);
					refresh = 1;
					break;
				case ORNAMENT+GUP :
					tool->percent = functions->ShiftEmbossedProp(tw,ORNAMENT,1,0);
					refresh = 1;
					break;
				case TRANSPOSE :
					tool->transposition = functions->DragEmbossedProp(tw,TRANSPOSE);
					refresh = 1;
					break;
				case TRANSPOSE+GDOWN :
					tool->transposition = functions->ShiftEmbossedProp(tw,TRANSPOSE,-1,0);
					refresh = 1;
					break;
				case TRANSPOSE+GUP :
					tool->transposition = functions->ShiftEmbossedProp(tw,TRANSPOSE,1,0);
					refresh = 1;
					break;
				case GATE :
					tool->gate = functions->DragEmbossedProp(tw,GATE);
					refresh = 1;
					break;
				case GATE+GDOWN :
					tool->gate = functions->ShiftEmbossedProp(tw,GATE,-1,0);
					refresh = 1;
					break;
				case GATE+GUP :
					tool->gate = functions->ShiftEmbossedProp(tw,GATE,1,0);
					refresh = 1;
					break;
				case IGNORE :
					tool->ignore = functions->DragEmbossedProp(tw,IGNORE);
					refresh = 1;
					break;
				case IGNORE+GDOWN :
					tool->ignore = functions->ShiftEmbossedProp(tw,IGNORE,-1,0);
					refresh = 1;
					break;
				case IGNORE+GUP :
					tool->ignore = functions->ShiftEmbossedProp(tw,IGNORE,1,0);
					refresh = 1;
					break;
				case DELAY :
					tool->delay = functions->DragEmbossedProp(tw,DELAY);
					refresh = 1;
					break;
				case DELAY+GDOWN :
					tool->delay = functions->ShiftEmbossedProp(tw,DELAY,-1,0);
					refresh = 1;
					break;
				case DELAY+GUP :
					tool->delay = functions->ShiftEmbossedProp(tw,DELAY,1,0);
					refresh = 1;
					break;

				case POPUP_IGNORE :
					code = functions->popupnote(tw,-1);
					if (code >= 0) {
						tool->ignore = 768 >> code;
					}
					refresh = 1;
					break;
				case POPUP_DELAY :
					code = functions->popupnote(tw,-1);
					if (code >= 0) {
						tool->delay = 768 >> code;
					}
					refresh = 1;
					break;

				default : break;
			}
		}
	}
	selectedtool = NULL;
	if (menu) functions->DeleteMenu(menu);
	functions->EmbossOff(tw,SCALE);
	functions->EmbossOff(tw,PLAY);
	functions->FatEmbossedPropOff(tw,ORNAMENT,ORNAMENT+GDOWN,ORNAMENT+GUP);
	functions->FatEmbossedPropOff(tw,TRANSPOSE,TRANSPOSE+GDOWN,TRANSPOSE+GUP);
	functions->FatEmbossedPropOff(tw,GATE,GATE+GDOWN,GATE+GUP);
	functions->FatEmbossedPropOff(tw,IGNORE,IGNORE+GDOWN,IGNORE+GUP);
	functions->FatEmbossedPropOff(tw,DELAY,DELAY+GDOWN,DELAY+GUP);
	for (code=C_ORNAMENT ; code <= C_TRANSPOSE ; code++ ) {
		functions->FatEmbossedPropOff(tw,code,code+GDOWN,code+GUP);
	}
	NewCloseEmbossedWindow(&tool->tool,nw);
}

/*==========================================================================
* savesize
*
*  INPUT: pointer to tool
* RETURN: bytes required to save the tool
*
* EFFECT: calculates the size less the noteon table (garbage list pointers
*         get saved)
*=========================================================================*/

static long __stdargs __saveds savesize(struct OrnaTool *tool)

{
	long size;

	size = TOOL_SIZE - (sizeof(long)*128) + 4;  // less on table + CInf_ID length
	size += sizeof(struct Event) * tool->count; // size of event lists
	if (tool->notes) size += 8;                 // add headers & counts
	if (tool->cchg) size += 8;
	if (tool->mtouch) size += 8;
	if (tool->ptouch) size += 8;
	if (tool->pbend) size += 8;
	return(size);
}

/*==========================================================================
* createtool
*
*  INPUT: none
* RETURN: pointer to Tool
*
* EFFECT: allocates and clears stuff
*=========================================================================*/

static struct Tool * __stdargs __saveds createtool(void)

{
	struct OrnaTool *tool;
	
	tool = (struct OrnaTool *)functions->myalloc(sizeof(struct OrnaTool),0);
	if (tool) {
		memset((void *)tool,0,TOOL_SIZE);
/*
		tool->flags = 0;
		tool->moflags = 0;
		tool->notes = NULL;
		tool->cchg = NULL;
		tool->mtouch = NULL;
		tool->ptouch = NULL;
		tool->pbend = NULL;
		tool->duration = 0;
		tool->count = 0;
*/
		ClearOnTable(tool)
	}
	return((struct Tool *)tool);
}


/*==========================================================================
* deletetool
*
*  INPUT: pointer to Tool
* RETURN: none
*
* EFFECT: cleans up & frees
*=========================================================================*/

static void __stdargs __saveds deletetool(struct OrnaTool *tool)

{
	freelists(tool);
	functions->myfree((char *)tool,TOOL_SIZE);
}

/*==========================================================================
* readlist
*
*  INPUT: file and event count
* RETURN: pointer to event list
*
* EFFECT: reads a list of events from a file
*=========================================================================*/

static struct Event * __regargs readlist(long file, long count)
{
	struct Event *event;
	struct Event *list = NULL;

	while (count) {
		event = (struct Event *)functions->myalloc(sizeof(struct Event),0);
		if (event) {
			memset((char *)event,0,sizeof(struct Event));
			functions->fastread(file,(char *)event,sizeof(struct Event));
			event->next = list;
			list = event;
			count--;
		}
		else {
			functions->fastseek(file,sizeof(struct Event),0);
		}
	}
	if (list) {
		list = functions->sorteventlist(list);
	}
	return(list);
}

/*==========================================================================
* writelist
*
*  INPUT: pointer to file, pointer to list, id
* RETURN: did or didn't
*
* EFFECT: writes the list to the file
*=========================================================================*/

static long __regargs writelist(long file, struct Event *list, long i)
{
	struct Event *event;
	long count;

	functions->fastwrite(file,(char *)&i,4); //id passed
#pragma msg 154 ignore push
	count = functions->List_Len((struct ListItem *)list);
#pragma msg 154 pop
	functions->fastwrite(file,(char *)&count,4); //event count
	for ( event=list ; event ; event = event->next ) {
		functions->fastwrite(file,(char *)event,sizeof(struct Event));
	}
	return(1);
}

/*==========================================================================
* loadtool
*
*  INPUT: file id, size of chunk to read
* RETURN: pointer to allocated tool, or NULL
*
* EFFECT: allocates and initializes a tool with data read from disk (remember
*         that the garbage list pointers get written to the file)
*
* !!NOTE: this function has been kludged to support loading an old song file
*         after adding the real-time controllers!
*=========================================================================*/

static struct Tool * __stdargs __saveds loadtool(long file, long size)
{
	struct OrnaTool *tool;
	long id;
	long count;
	long r;

	if (tool = (struct OrnaTool *)functions->myalloc(TOOL_SIZE,0)) {
		count = TOOL_SIZE - (sizeof(long) * 128) - CINFO_LENGTH;
		inittool(tool); //ensure reasonable values for stuff not saved in old version
//		memset((void *)tool,0,TOOL_SIZE);
//		ClearOnTable(tool);
		functions->fastread(file,(char *)tool,count);
		size -= count;
		while (size > 0) {
			r = functions->fastread(file,(char *)&id,4);
			size -= r;
			if (r != 4) continue;
			if (id != CInf_ID) { // no byte count for only this ID
				r = functions->fastread(file,(char *)&count,4);
				size -= r;
				if (r != 4) continue;
			}
			switch (id) {
				case Note_ID :
					tool->notes = readlist(file,count);
					break;
				case Bend_ID :
					tool->pbend = readlist(file,count);
					break;
				case CChg_ID :
					tool->cchg = readlist(file,count);
					break;
				case MTch_ID :
					tool->mtouch = readlist(file,count);
					break;
				case PTch_ID :
					tool->ptouch = readlist(file,count);
					break;
				case CInf_ID : //sort of kludgey way to not rewrite this function
					if (size >= CINFO_LENGTH) {
						functions->fastread(file,(char *)&tool->cinfo,CINFO_LENGTH);
						size -= CINFO_LENGTH;
						continue;
					}
					break;
#if DBUG
				default : // the file is fucked
dprintf("   STRANGE PHENOMENON\n");
					break;
#endif
			}
			size = size - (count * sizeof(struct Event));
		}
	}
	if (size) functions->fastseek(file,size,0); //skip leftovers
	return((struct Tool *)tool);
}

static struct ToolMaster master;

/*==========================================================================
* savetool
*
*  INPUT: file id, pointer to tool to save
* RETURN: 0 if successful; 1 otherwise
*
* EFFECT: writes tool data to a file (including the garbage list pointers)
*=========================================================================*/

static long __stdargs __saveds savetool(long file, struct OrnaTool *tool)

{
	long size;

	functions->fastwrite(file,(char *)&master.toolid,4);
	size = savesize(tool);
	functions->fastwrite(file,(char *)&size,4);
	functions->fastwrite(file,(char *)tool,TOOL_SIZE-(sizeof(long)*128)-CINFO_LENGTH);
	if (tool->notes) {
		writelist(file,tool->notes,Note_ID);
	}
	if (tool->pbend) {
		writelist(file,tool->pbend,Bend_ID);
	}
	if (tool->cchg) {
		writelist(file,tool->cchg,CChg_ID);
	}
	if (tool->mtouch) {
		writelist(file,tool->mtouch,MTch_ID);
	}
	if (tool->ptouch) {
		writelist(file,tool->ptouch,PTch_ID);
	}
	 //kludge to add controller stuff & still support old tools saved previously
	size = CInf_ID;
	functions->fastwrite(file,(char *)&size,4);
	functions->fastwrite(file,(char *)&tool->cinfo,CINFO_LENGTH);
	return(0);
}

/****************************************/
static void __stdargs __saveds transportcode(command, time, end)

register long command;
register long time;
register long end;

{
	register struct OrnaTool *tool;
	register struct Track *track;

	if (command == TC_STOP) {
		track = (struct Track *)functions->tracklist;
		for ( ; track ; track = track->next ) {
			tool = (struct OrnaTool *)track->toollist;
			for ( ; tool ; tool = (struct OrnaTool *)tool->tool.next ) {
				if (tool->tool.toolid == TOOLID) {
					ClearOnTable(tool);
				}
			}
		}
	}
}
/*******************************/
static void __stdargs __saveds removetool(void)
{
	functions->removetransport(transportcode);
}


/*==========================================================================
* inittoolmaster
*
*  INPUT: none
* RETURN: pointer to ToolMaster
*
* EFFECT: initializes.
*=========================================================================*/

struct ToolMaster * __stdargs __saveds inittoolmaster(void)

{
	memset((char *)&master,0,sizeof(struct ToolMaster));
	master.toolsize = TOOL_SIZE;
	master.toolid = TOOLID;
	master.edittool = edittool;
	master.processevent = processevent;
	master.processlist = processlist;
	master.createtool = createtool;
	master.deletetool = deletetool;
	master.savesize = savesize;
	master.savetool = savetool;
	master.loadtool = loadtool;
	master.removetool = removetool;
	master.image = &IconDown;
	master.upimage = &IconUp;
	strcpy(master.name,WTITLE);
	master.tooltype = TOOL_NORMAL | TOOL_BRANCHOUT;
	functions->installtransport((void *)transportcode);
	return(&master);
}
